from onepassword_tools.lib.OnePasswordDatabaseItem import OnePasswordDatabaseItem
from onepassword_tools.lib.NewItemCommand import NewItemCommand, new_item_command_options, new_item_command_password
import click


@click.command()
@new_item_command_options
@new_item_command_password
@click.option('--database', help='Database name', prompt=False, required=True)
@click.option('--hostname', help='Host where the account is created', prompt=True, required=True)
@click.option('--port', help='Database port', prompt=False, required=False, default='')
@click.option('--username', help='Account username', prompt=True, required=True)
def new_database_account(hostname, database, port, username, account, notes, password, password_length, return_field, title, vault, do_not_ask_credentials):
    """Create a new Database item in 1Password with the given credentials."""
    NewDatabaseAccount(**locals()).run()


class NewDatabaseAccount(NewItemCommand):

    database: str = None
    hostname: str = None
    port: str = None
    titleTemplate = 'USER {username} DB {database} ON {hostname}'
    username: str = None
    onePasswordItemClass = OnePasswordDatabaseItem



from onepassword_tools.lib.OnePasswordItem import OnePasswordItem


class OnePasswordSSHKeyItem(OnePasswordItem):

    from_user: str = None
    from_host: str = None
    to_host: str = None
    to_host_abbreviated: str = None
    to_ip: str = None
    to_user: str = None
    passphrase: str = None
    public_key: str = None
    private_key: str = None
    item_type: str = 'Server'
    title: str = None
    notes: str = ''
    tags: [] = ['Clef SSH']
    url: str = None
    username: str
    to_port: str = None

    def __getitem__(self, item):
        if item == 'url':
            return self._get_url()
        else:
            return super().__getitem__(item)

    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            if hasattr(self, key):
                self.__setattr__(key, value)

    def _get_url(self):
        if self.to_host and self.to_user:
            return "ssh://%s@%s" % (self.to_user, self.to_host)
        else:
            return None

    def _request_object(self):
        """
        Return the dict object with the request sent to 1Password
        :return: dict
        """
        section = {
            "title": "",
            "name": "Section_%s" % self.opu.generate_op_section_uuid()
        }

        return {
            "category": "SERVER",
            "fields": [
                {
                    "id": "notesPlain",
                    "type": "STRING",
                    "purpose": "NOTES",
                    "label": "notesPlain",
                    "value": self.get('notes'),
                },
                {
                    "type": "concealed",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('passphrase'),
                    "label": "Passphrase",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('to_host'),
                    "label": "Hostname",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('to_ip'),
                    "label": "IP",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('public_key'),
                    "label": "Public Key",
                    "section": section
                },
                {
                    "id": "private_key",
                    "type": "UNKNOWN",
                    "value": self.get('private_key'),
                    "label": "private key",
                    "section": section
                },
                {
                    "type": "string",
                    "a": {
                        "multiline": "yes"
                    },
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('private_key'),
                    "label": "Private Key",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('url'),
                    "label": "url",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": str(self.get('to_port')),
                    "label": "Port",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('to_host_abbreviated'),
                    "label": "Hostname abbreviated",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('to_user'),
                    "label": "Remote user",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('from_user'),
                    "label": "Local user",
                    "section": section
                },
                {
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('from_host'),
                    "label": "Local host",
                    "section": section
                }
            ],
            "sections": [
                section
            ]
        }



from onepassword_tools.lib.OnePasswordItem import OnePasswordItem


class OnePasswordServerItem(OnePasswordItem):

    hostname: str = None
    item_type: str = 'Server'
    notes: str = None
    password: str = None
    tags: [] = ['Compte utilisateur']
    url: str = None
    username: str = None

    def __getitem__(self, item):
        if item == 'url':
            return self._get_url()
        else:
            return super().__getitem__(item)

    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            if hasattr(self, key):
                self.__setattr__(key, value)

    def _get_url(self):
        if self.username and self.hostname:
            return "ssh://%s@%s" % (self.username, self.hostname)
        elif self.hostname:
            return "ssh://%s" % self.hostname
        else:
            return None

    def _request_object(self):
        """
        Return the dict object with the request sent to 1Password
        :return: dict
        """
        main_section_uuid = {
            "title": "",
            "name": "Section_%s" % self.opu.generate_op_section_uuid()
        }
        admin_section_uuid = {
            "title": "",
            "name": "Section_%s" % self.opu.generate_op_section_uuid()
        }

        hosting_provider_section_uuid = {
            "title": "",
            "name": "Section_%s" % self.opu.generate_op_section_uuid()
        }

        return {
            "category": "SERVER",
            "fields": [
                {
                    "id": "notesPlain",
                    "type": "STRING",
                    "purpose": "NOTES",
                    "label": "notesPlain",
                    "value": self.get('notes')
                },
                {
                    "inputTraits": {
                        "keyboard": "URL"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('url'),
                    "label": "URL",
                    "section": main_section_uuid
                },
                {
                    "inputTraits": {
                        "autocapitalization": "none",
                        "autocorrection": "no"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('username'),
                    "label": "username",
                    "section": main_section_uuid
                },
                {
                    "type": "concealed",
                    "id": self.opu.generate_op_field_uuid(),
                    "value": self.get('password'),
                    "label": "password",
                    "section": main_section_uuid
                },
                {
                    "inputTraits": {
                        "keyboard": "URL"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "admin console URL",
                    "section": admin_section_uuid
                },
                {
                    "inputTraits": {
                        "autocapitalization": "none",
                        "autocorrection": "no"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "admin console username",
                    "section": admin_section_uuid
                },
                {
                    "type": "concealed",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "console password",
                    "section": admin_section_uuid
                },
                {
                    "inputTraits": {
                        "autocapitalization": "Words"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "name",
                    "section": hosting_provider_section_uuid
                },
                {
                    "inputTraits": {
                        "keyboard": "URL"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "website",
                    "section": hosting_provider_section_uuid
                },
                {
                    "inputTraits": {
                        "keyboard": "URL"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "support URL",
                    "section": hosting_provider_section_uuid
                },
                {
                    "inputTraits": {
                        "keyboard": "NamePhonePad"
                    },
                    "type": "string",
                    "id": self.opu.generate_op_field_uuid(),
                    "label": "support phone",
                    "section": hosting_provider_section_uuid
                }
            ],
            "sections": [
                main_section_uuid,
                admin_section_uuid,
                hosting_provider_section_uuid
            ]
        }




from onepassword_tools.lib.OnePasswordServerItem import OnePasswordServerItem
from onepassword_tools.lib.NewItemCommand import NewItemCommand, new_item_command_options, new_item_command_password
import click


@click.command()
@new_item_command_options
@new_item_command_password
@click.option('--hostname', help='Host where the account is created', prompt=True, required=True)
@click.option('--username', help='Account username', prompt=True, required=True)
def new_server_account(hostname, username, password, password_length, vault, return_field, account, title, notes, do_not_ask_credentials):
    """Create a new Server item in 1Password with the given credentials."""
    NewServerAccount(**locals()).run()


class NewServerAccount(NewItemCommand):

    hostname: str = None
    onePasswordItemClass = OnePasswordServerItem
    titleTemplate = 'USER {username} ON {hostname}'
    username: str = None





from onepassword_tools.lib.OnePasswordUtils import OnePasswordUtils
from onepassword_tools.lib.MiscUtils import remove_null_value_keys_in_dict
import uuid
from abc import ABCMeta, abstractmethod


class OnePasswordItem:
    __metaclass__ = ABCMeta

    createCustomUUID: bool = True

    @property
    @abstractmethod
    def item_type(self) -> str:
        pass

    def __init__(self):
        self.opu = OnePasswordUtils()

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, item):
        if hasattr(self, item):
            return getattr(self, item)
        else:
            return None

    @abstractmethod
    def _request_object(self):
        """
        Static dictionnary reprensenting the request sent to 1Password to create entry
        :return: dict
        """
        return

    def get(self, item):
        if item == 'notes' and getattr(self, item) is None:
            return ""
        else:
            return self[item]

    def get_uuid_section(self, section_uuid):
        return {
            "title": "",
            "name": section_uuid,
        }

    def get_uuid_field(self, section_uuid):
        return {
            "label": "UUID",
            "value": str(uuid.uuid4()),
            "type": "STRING",
            "id": self.opu.generate_op_field_uuid(),
            "section": self.get_uuid_section(section_uuid)
        }

    def get_request_object(self):
        request = self._request_object()

        if self.createCustomUUID:
            section_uuid = "Section_%s" % self.opu.generate_op_section_uuid()
            request['sections'].append(self.get_uuid_section(section_uuid))
            request['fields'].append(self.get_uuid_field(section_uuid))

        request = remove_null_value_keys_in_dict(request)

        return request



from onepassword_tools.lib.ClickUtils import ClickUtils
from onepassword_tools.lib.ConfigFile import ConfigFile
from onepassword_tools.lib.Log import Log
from onepassword_tools.lib.MiscUtils import is_uuid
import json
import os
import secrets
import string
import subprocess
import sys
if os.environ.get('USE_LOCAL'):
    sys.path.insert(0, os.path.dirname(os.path.abspath(__file__)) + "/../../../onepassword-local-search")
from onepassword_local_search.models.Item import Item
from onepassword_local_search.OnePassword import OnePassword
from onepassword_local_search.exceptions.ManagedException import ManagedException
import re
import tempfile


class OnePasswordUtils:

    onePassword: OnePassword
    suggestions: [] = []

    def __init__(self):
        self.onePassword = OnePassword()
        self.config = ConfigFile()

    @staticmethod
    def _communicate(subproc, inputstr=None):
        """
        Encode the input given to the subprocess if any
        :param subproc:
        :param inputstr:
        :return: A tuple of (stdout, stderr)
        """
        if inputstr is None:
            return subproc.communicate()
        if not inputstr:
            inputstr = ''
        return subproc.communicate(inputstr.encode())

    def _authenticate(self, shorthand=''):
        """
        Authenticate over 1Password and register the session key in the environment variable
        :return: Nothing
        """
        try:

            sessionKey = subprocess.check_output([
                'op',
                'signin',
                '--account', self.onePassword.configFileService.get_account_uuid_from_account_shorthand(shorthand),
                '--raw'
            ]) \
                .decode('utf-8') \
                .replace('\n', '')
            os.environ[
                'OP_SESSION_' + self.onePassword.configFileService.get_user_uuid_from_account_shorthand(shorthand)
            ] = sessionKey

        except subprocess.CalledProcessError:
            ClickUtils.error('Failed to authenticate. You may have written the wrong password ?')
            sys.exit(1)

    def authenticate(self, shorthand=None):
        if shorthand is None:
            shorthands = self.config.get_section('accounts')
            if len(shorthands) > 0:
                for shorthand in shorthands:
                    self._authenticate(shorthand)
            else:
                self._authenticate(self.onePassword.configFileService.get_latest_signin())
        else:
            self._authenticate(shorthand)
        self.onePassword.accountService.init()

    def create_item(self, request_object, category, title, tags=None, url='', vault='', account=''):
        if tags is None:
            tags = []
        try:
            Log.debug(request_object, 'request data')
            data = json.dumps(request_object)
            with tempfile.NamedTemporaryFile() as requestFile:
                requestFile.write(data.encode('utf-8'))
                requestFile.flush()

                command = 'item create --format json --template %s' % (requestFile.name)
                if title and title != '':
                    command += ' --title="%s"' % title
                if type(tags).__name__ != 'list':
                    tags = [tags]
                if tags and len(tags) > 0:
                    command += ' --tags="%s"' % ','.join(tags)
                if vault and vault != '':
                    command += ' --vault="%s"' % vault
                if account and account != '':
                    command += ' --account=%s' % account

                Log.debug(command, 'op command : %s', command)
                rc, output, error = self.op_cli(command)
                requestFile.close()

                if output is not None and len(output) > 1:
                    created_item = json.loads(output.replace('\n', ''))
                    if not created_item.get('id'):
                        Log.error(error)
                        raise Exception("Error while creating onepassword entry")
                    else:
                        created_item['request_object'] = request_object
                        Log.debug(created_item)
                        return created_item
                else:
                    Log.error(error)
                    raise Exception("Error while creating onepassword entry")

        except Exception:
            if requestFile:
                requestFile.close()
            raise Exception("Error while creating onepassword entry")

    def is_authenticated(self, account=None, check_mode=None) -> bool:
        """
        Check if we are authenticated over 1Password by decrypting local keys
        and trying to grab a template remotely
        :return: bool
        """
        if check_mode is None or check_mode == 'local':
            if not self.onePassword.is_authenticated():
                return False

        if check_mode is None or check_mode == 'remote':
            if account is None:
                accounts = self.config.get_section('accounts')
                if len(accounts) > 0:
                    for account_ in accounts:
                        if not self.op_cli('account get --account="%s"' % account_)[0] == 0:
                            return False
                else:
                    if not self.op_cli('account get')[0] == 0:
                        return False
            else:
                if not self.op_cli('account get --account="%s"' % account)[0] == 0:
                    return False
        return True

    def generate_op_field_uuid(self):
        return self.generate_op_uuid(26)

    def generate_op_section_uuid(self):
        return self.generate_op_uuid(29)

    @staticmethod
    def generate_op_uuid(length=26):
        return ''.join(secrets.choice(string.ascii_lowercase) for _ in range(length))

    def get_alias(self, search):
        """
        Search if the input string is aliases to an uuid in the config file
        :param search:
        :return:
        """
        if self.config.config_key_exists('aliases', search):
            return self.config.config['aliases'][search]
        else:
            return search

    def op_cli(self, cmd, inputstr=None):
        """
        Call the 1Password cli (op)
        :param cmd: the op arguments and options
        :param inputstr: the input given to the command
        :return: A tuple of (returncode, stdout, stderr)
        """
        e = os.environ.copy()
        p = subprocess.Popen('op ' + cmd,
                             bufsize=0,
                             close_fds=True,
                             env=e,
                             shell=True,
                             stdin=subprocess.PIPE,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE
                             )
        try:
            out, err = self._communicate(p, inputstr=inputstr)
        except subprocess.CalledProcessError:
            out = ''
            err = ''
        rc = p.returncode
        return rc, out.decode('utf-8'), err.decode('utf-8')

    def search_item_uuid_by_title(self, search, search_operator='AND'):
        """
        Search an item by title
        :param search:
        :return: Item if only one match, None either
        """
        items = self.onePassword.get_items(search, search_operator)
        if len(items) == 1:
            return self.onePassword.get(items[0].uuid, output=False)
        else:
            return items

    def try_to_grab_item(self, search, search_operator='AND') -> Item:
        """
        First try to grab by using 1Password or custom UUID, then fallback to
        search item by title.
        :param search:
        :return: Item or None
        """
        search = [self.get_alias(search[0])]
        try:
            item = self.onePassword.get(search[0], output=False)
        except ManagedException:
            item = self.search_item_uuid_by_title(search, search_operator)
        return item
